from fpylll import LLL, BKZ, IntegerMatrix, GSO
from fpylll.algorithms.bkz2 import BKZReduction as BKZ2
from sage.all import *
from sage.crypto.lwe import samples, LWE
import time
import multiprocessing
from numpy import zeros, nan

from tools import CenteredBinomialDistibution, SVPreduction, calculate_t, score, makenumpy
from plots import plot_lll, plot_bkz

def nistlike(n, q):
    """
        Given n, q, generate a NIST-like LWE instance

        args:
        n (int): dimension
        q (int): modulus

        returns:
        A (matrix): matrix
        t (array): target vector
        s (array): secret vector
        e (array): error vector
    """
    m=1
    DB = CenteredBinomialDistibution(2)
    lwe_instance = LWE(n=n, q=q, D=DB, secret_dist='noise', m=m)
    s = lwe_instance._LWE__s
    s = array([int(x) for x in s])
    s = array([int(x) if x<= q//2 else int(x-q) for x in s], dtype=int)
    a = samples(m, n, lwe_instance, balanced=True)
    a = a[0][0]
    assert all(-q//2 <=x <= q // 2 for x in a)
    A = zeros((n, n), dtype=int)
    for i in range(n):
        for j in range(n):
            if j > i:
                A[i][j] = int(-a[i-j])
            else:
                A[i][j] = int(a[i-j])
    e = zeros(n, dtype=int)
    for i in range(n):
        e[i] = int(DB())
    t = calculate_t(A, s, e, q)
    return A, t, s, e

def bkz_with_blocksize(instance):
    A = instance[0]
    t = instance[1]
    s = instance[2]
    e = instance[3]
    q = instance[4]
    b = instance[5]
    B = SVPreduction(A, t, q)
    params = BKZ.Param(b, strategies=BKZ.DEFAULT_STRATEGY, flags=BKZ.AUTO_ABORT)
    bkz = BKZ2(GSO.Mat(B, float_type='mpfr'))
    try:
        bkz(params)
    except:
        return None
    B_np = makenumpy(B)
    s_best = B_np[0][:n]
    e_best = B_np[0][n:-1]
    sc = score(t, s_best, e_best, A, s, p)
    if sc:
        return True
    else: return False

def time_bkz(task_function, *args, **kwargs):
    start_time = time.time()
    result = task_function(*args)
    end_time = time.time()
    execution_time = end_time - start_time
    return result, execution_time


# Run LLL
tries = 128
pp = [3, 17, 71, 277, 401, 521, 1031, 3329]
nn = range(8, 81, 8)
for p in pp:
    x = zeros(len(nn))
    k = 0
    for n in nn:
        for _ in range(tries):
            A, t, s, e = nistlike(n, p)
            B = SVPreduction(A.copy(), t, p)
            L = LLL.reduction(B)
            B_np = makenumpy(L)
            s_best = B_np[0][:n]
            e_best = B_np[0][n:2*n]
            sc = score(t, s_best, e_best, A, s, p)
            if sc:
                x[k] += 1
        x[k] = x[k] / tries
        output = open('lll_results.txt', 'a')
        output.write(f'{p} {n} {x[k-1]}\n')
        output.close()
        k += 1

plot_lll()



# Run BKZ
tries = 256
pp = [71, 401, 3329]
nn = range(8, 113, 8)
blocks = [8, 16, 24, 32, 40, 48, 56]

for p in pp:
    for n in nn:
        inst = []
        for _ in range(tries):
            A, t, s, e = nistlike(n, p)
            i = (A, t, s, e, p)
            inst.append(i)
        for b in blocks:
            b_inst = [(i[0].copy(), i[1].copy(), i[2].copy(), i[3].copy(), i[4].copy(), b) for i in inst]
            with multiprocessing.Pool() as pool:
                result = pool.starmap(time_bkz, [(bkz_with_blocksize, i) for i in b_inst])
            res = [r for r in result if r[0] != None]
            if len(res) == 0:
                # fpylll infinite babei loop error for all instances
                tt = nan
                prob = 0
            else:
                tt = sum(r[1] for r in res)/len(res)
                prob = sum(r[0] for r in res)/len(res)
            output = open('bkz_results.txt', 'a')
            output.write(f"{n} {p} {b} {prob} {tt} {len(res)}\n")
            output.close()

plot_bkz()
