from fpylll import IntegerMatrix
from numpy import zeros, array, matrix, dot, arange, nan
from numpy.linalg import norm
import random
import multiprocessing
import time

class CenteredBinomialDistibution:

    def __init__(self, n):
        self.n = n
    def __call__(self):
        x, y = 0, 0
        for k in range(self.n):
            x += random.randint(0, 1)
            y += random.randint(0, 1)
        return x - y

def SVPreduction(A, t, q):
    """
        Given A, t, q, return the SVP-reduced basis using an embedding technique
        B = [I_n | -A^T   | 0
             0   |  qI_n  | 0
             0   |  t     | 1]

        args:
        A (matrix): matrix
        t (array): target vector
        q (int): modulus

        returns:
        B (matrix): SVP-reduced basis
    """
    d, n = A.shape
    B = IntegerMatrix(n+d+1,n+d+1)
    for i in range(n):
        for j in range(n, n+d):
            B[i,j] = -A[j - n][i]

    for i in range(n):
        B[i,i] = 1

    for i in range(n,n+d):
        B[i,i] = q

    B[-1,-1] = 1

    for i in range(n, n+d):
        B[-1,i] = t[i-n]

    return B

def calculate_t(A:matrix, s:array, e:array, q:int) -> array:
    """
        Given A, s, e, q, calculate t = (A*s + e) mod q

        args:
        A (matrix): matrix
        s (array): secret vector
        e (array): error vector
        q (int): modulus
        returns:
        array t
    """
    t = array([(dot(A[j], s) + e[j]) % q for j in range(len(A))])
    for i in range(len(t)):
        t[i] = t[i] if t[i] <= q//2 else t[i]-q
    return t

def makenumpy(B):
    """
        Convert fpylll integer matrix to numpy array

        args:
        B (fpylll integer matrix): matrix

        returns:
        A (numpy array): matrix
    """
    A = zeros((B.nrows, B.ncols), dtype=int)
    for i in range(B.nrows):
        for j in range(B.ncols):
            A[i][j] = int(B[i,j])
    return A

def score(t, ss, ee, A, s, p):
    """
        Given t, s, e, A, p, calculate the score
        Calculate t' = (A*s + e) mod p
        Returns ||t-t'||_1

        args:
        t (array): target vector
        ss (array): secret vector
        ee (array): error vector
        A (matrix): matrix
        p (int): modulus

        returns:
        score (int): score
    """
    tt = calculate_t(A, ss, ee, p)
    for k in range(len(t)):
        if t[k] >= p//2:
            t[k] = t[k] - p
        if tt[k] >= p//2:
            tt[k] = tt[k] - p
    return t==tt
